@@ -2,8 +2,8 @@
 
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
-    2006, 2007, 2008, 2009, 2010, 2011 Massachusetts Institute of
-    Technology
+    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
+    2017, 2018 Massachusetts Institute of Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -37,26 +37,40 @@ USA.
        packages
        extensions
        loads
-       (map (lambda (pathname)
-	      (cons
-	       (->namestring pathname)
-	       (let ((pathname
-		      (package-set-pathname
-		       (merge-pathnames pathname model-pathname)
-		       os-type)))
-		 (if (file-exists? pathname)
-		     (let ((contents (fasload pathname #t)))
-		       (if (package-file? contents)
-			   contents
-			   (begin
-			     (warn "Malformed package-description file:"
-				   pathname)
-			     #f)))
-		     (begin
-		       (warn "Can't find package-description file:" pathname)
-		       #f)))))
+       (map (lambda (name)
+	      (let ((pathname (find-global-definitions name model-pathname
+						       os-type)))
+		(and pathname
+		     (cons (->namestring pathname)
+			   (let ((contents (fasload pathname #t)))
+			     (if (package-file? contents)
+				 contents
+				 (begin
+				   (warn "Malformed package-description file:"
+					 pathname)
+				   #f)))))))
 	    globals)
        model-pathname))))
+
+(define (find-global-definitions name model-pathname os-type)
+  (let* ((filename (->pathname
+		    (cond ((symbol? name) (symbol->string name))
+			  ((string? name) name)
+			  (else (error "Not a globals name:" name)))))
+	 (pkd (package-set-pathname filename os-type)))
+    (or
+     (if (symbol? name)
+	 (let ((pathname (ignore-errors
+			  (lambda ()
+			    (system-library-pathname pkd)))))
+	   (and (not (condition? pathname))
+		pathname))
+	 (let ((pathname (merge-pathnames pkd model-pathname)))
+	   (and (file-exists? pathname)
+		pathname)))
+     (begin
+       (warn "Could not find global definitions:" pkd)
+       #f))))
 
 (define (sort-descriptions descriptions)
   (letrec
@@ -66,7 +80,7 @@ USA.
 	      (let ((description (car descriptions))
 		    (descriptions (cdr descriptions)))
 		(case (car description)
-		  ((DEFINE-PACKAGE)
+		  ((define-package)
 		   (loop descriptions
 			 (cons (cdr description) packages)
 			 extensions
@@ -74,7 +88,7 @@ USA.
 			     (cons (cdr description) loads)
 			     loads)
 			 globals))
-		  ((EXTEND-PACKAGE)
+		  ((extend-package)
 		   (loop descriptions
 			 packages
 			 (cons (cdr description) extensions)
@@ -82,13 +96,13 @@ USA.
 			     (cons (cdr description) loads)
 			     loads)
 			 globals))
-		  ((GLOBAL-DEFINITIONS)
+		  ((global-definitions)
 		   (loop descriptions
 			 packages
 			 extensions
 			 loads
 			 (append! (reverse (cdr description)) globals)))
-		  ((NESTED-DESCRIPTIONS)
+		  ((nested-descriptions)
 		   (receive (packages extensions loads globals)
 		       (loop (cdr description)
 			     packages
@@ -203,11 +217,11 @@ USA.
 		    (make-reference primitive-package
 				    (primitive-procedure-name name)
 				    expression))
-		   ((access? name)
-		    (if (eq? (access-environment name)
+		   ((scode-access? name)
+		    (if (eq? (scode-access-environment name)
 			     system-global-environment)
 			(make-reference root-package
-					(access-name name)
+					(scode-access-name name)
 					expression)
 			(warn "Non-root access" (unsyntax name))))
 		   (else
@@ -219,7 +233,7 @@ USA.
 (define (resolve-references! pmodel)
   (for-each (lambda (package)
 	      (for-each resolve-reference!
-			(package/sorted-references package)))
+			(package/references package)))
 	    (pmodel/packages pmodel)))
 
 (define (resolve-reference! reference)
@@ -255,43 +269,43 @@ USA.
 		  (list? (cdr expression))))
 	(lose))
     (case (car expression)
-      ((DEFINE-PACKAGE)
-       (cons 'DEFINE-PACKAGE
+      ((define-package)
+       (cons 'define-package
 	     (parse-package-definition (parse-name (cadr expression))
 				       (cddr expression))))
-      ((EXTEND-PACKAGE)
-       (cons 'EXTEND-PACKAGE
+      ((extend-package)
+       (cons 'extend-package
 	     (parse-package-extension (parse-name (cadr expression))
 				      (cddr expression))))
-      ((GLOBAL-DEFINITIONS)
+      ((global-definitions)
        (let ((filenames (cdr expression)))
-	 (if (not (for-all? filenames string?))
+	 (if (not (every (lambda (f) (or (string? f) (symbol? f))) filenames))
 	     (lose))
-	 (cons 'GLOBAL-DEFINITIONS (map parse-filename filenames))))
-      ((OS-TYPE-CASE)
+	 (cons 'global-definitions filenames)))
+      ((os-type-case)
        (if (not (and (list? (cdr expression))
-		     (for-all? (cdr expression)
-		       (lambda (clause)
-			 (and (or (eq? 'ELSE (car clause))
-				  (and (list? (car clause))
-				       (for-all? (car clause) symbol?)))
-			      (list? (cdr clause)))))))
+		     (every (lambda (clause)
+			      (and (or (eq? 'else (car clause))
+				       (and (list? (car clause))
+					    (every symbol? (car clause))))
+				   (list? (cdr clause))))
+			    (cdr expression))))
 	   (lose))
-       (cons 'NESTED-DESCRIPTIONS
+       (cons 'nested-descriptions
 	     (let loop ((clauses (cdr expression)))
 	       (cond ((null? clauses)
 		      '())
-		     ((or (eq? 'ELSE (caar clauses))
+		     ((or (eq? 'else (caar clauses))
 			  (memq os-type (caar clauses)))
 		      (parse-package-expressions (cdar clauses)
 						 pathname
 						 os-type))
 		     (else
 		      (loop (cdr clauses)))))))
-      ((INCLUDE)
-       (cons 'NESTED-DESCRIPTIONS
+      ((include)
+       (cons 'nested-descriptions
 	     (let ((filenames (cdr expression)))
-	       (if (not (for-all? filenames string?))
+	       (if (not (every string? filenames))
 		   (lose))
 	       (append-map (lambda (filename)
 			     (read-and-parse-model
@@ -305,26 +319,28 @@ USA.
 (define (parse-package-definition name options)
   (check-package-options options)
   (receive (parent options)
-      (let ((option (assq 'PARENT options)))
+      (let ((option (assq 'parent options)))
 	(if option
 	    (let ((options (delq option options)))
 	      (if (not (and (pair? (cdr option))
 			    (null? (cddr option))))
 		  (error "Ill-formed PARENT option:" option))
-	      (if (assq 'PARENT options)
+	      (if (assq 'parent options)
 		  (error "Multiple PARENT options."))
-	      (values (parse-name (cadr option)) options))
-	    (values 'NONE options)))
+	      (values (and (cadr option)
+			   (parse-name (cadr option)))
+		      options))
+	    (values 'none options)))
     (let ((package (make-package-description name parent)))
       (process-package-options package options)
       package)))
 
 (define (parse-package-extension name options)
   (check-package-options options)
-  (let ((option (assq 'PARENT options)))
+  (let ((option (assq 'parent options)))
     (if option
 	(error "PARENT option illegal in package extension:" option)))
-  (let ((package (make-package-description name 'NONE)))
+  (let ((package (make-package-description name 'none)))
     (process-package-options package options)
     package))
 
@@ -341,34 +357,55 @@ USA.
 (define (process-package-options package options)
   (for-each (lambda (option)
 	      (case (car option)
-		((FILES)
+		((files)
 		 (set-package-description/file-cases!
 		  package
 		  (append! (package-description/file-cases package)
 			   (list (parse-filenames (cdr option))))))
-		((FILE-CASE)
+		((file-case)
 		 (set-package-description/file-cases!
 		  package
 		  (append! (package-description/file-cases package)
 			   (list (parse-file-case (cdr option))))))
-		((EXPORT)
+		((export)
+		 (let ((export
+			(cond ((and (pair? (cdr option))
+				    (eq? 'deprecated (cadr option)))
+			       (parse-import/export (cddr option) #t))
+			      ;; 9.2 compatibility
+			      ((and (pair? (cdr option))
+				    (pair? (cddr option))
+				    (symbol? (caddr option))
+				    (string-prefix-ci?
+				     "deprecated:"
+				     (symbol->string (caddr option))))
+			       (parse-import/export (cons (cadr option)
+							  (cdddr option))
+						    #t))
+			      (else
+			       (parse-import/export (cdr option) #f)))))
+		   (set-package-description/exports!
+		    package
+		    (append! (package-description/exports package)
+			     (list export)))))
+		((export-deprecated)
 		 (set-package-description/exports!
 		  package
 		  (append! (package-description/exports package)
-			   (list (parse-import/export (cdr option))))))
-		((IMPORT)
+			   (list (parse-import/export (cdr option) #t)))))
+		((import)
 		 (set-package-description/imports!
 		  package
 		  (append! (package-description/imports package)
-			   (list (parse-import/export (cdr option))))))
-		((INITIALIZATION)
+			   (list (parse-import/export (cdr option) #f)))))
+		((initialization)
 		 (let ((initialization (parse-initialization (cdr option))))
 		   (if initialization
 		       (set-package-description/initializations!
 			package
 			(append! (package-description/initializations package)
 				 (list initialization))))))
-		((FINALIZATION)
+		((finalization)
 		 (let ((finalization (parse-initialization (cdr option))))
 		   (if finalization
 		       (set-package-description/finalizations!
@@ -387,7 +424,7 @@ USA.
 (define (parse-filenames filenames)
   (if (not (check-list filenames string?))
       (error "illegal filenames" filenames))
-  (list #F (cons 'ELSE (map parse-filename filenames))))
+  (list #f (cons 'else (map parse-filename filenames))))
 
 (define (parse-file-case file-case)
   (if (not (and (pair? file-case)
@@ -395,7 +432,7 @@ USA.
 		(check-list (cdr file-case)
 		  (lambda (clause)
 		    (and (pair? clause)
-			 (or (eq? 'ELSE (car clause))
+			 (or (eq? 'else (car clause))
 			     (check-list (car clause) symbol?))
 			 (check-list (cdr clause) string?))))))
       (error "Illegal file-case" file-case))
@@ -415,7 +452,7 @@ USA.
 	(warn "Illegal initialization/finalization:" initialization)
 	#f)))
 
-(define (parse-import/export object)
+(define (parse-import/export object deprecated?)
   (if (not (and (pair? object)
 		(check-list (cdr object)
 			    (lambda (item)
@@ -429,20 +466,20 @@ USA.
   (cons (parse-name (car object))
 	(map (lambda (entry)
 	       (if (pair? entry)
-		   (cons (car entry) (cadr entry))
-		   (cons entry entry)))
+		   (vector (car entry) (cadr entry) deprecated?)
+		   (vector entry entry deprecated?)))
 	     (cdr object))))
 
 (define (check-list items predicate)
   (and (list? items)
-       (for-all? items predicate)))
+       (every predicate items)))
 
 ;;;; Packages
 
 (define (descriptions->pmodel descriptions extensions loads globals pathname)
   (let ((packages
 	 (map (lambda (description)
-		(make-package (package-description/name description) 'UNKNOWN))
+		(make-package (package-description/name description) 'unknown))
 	      descriptions))
 	(extra-packages '()))
     (let ((root-package
@@ -457,13 +494,13 @@ USA.
 		       (begin
 			 (if (not intern?)
 			     (warn "Unknown package name:" name))
-			 (let ((package (make-package name 'UNKNOWN)))
+			 (let ((package (make-package name 'unknown)))
 			   (set! extra-packages
 				 (cons package extra-packages))
 			   package)))))))
 	;; GLOBALS is a list of the bindings supplied externally.
 	(for-each (lambda (global)
-		    (if (cdr global)
+		    (if (and global (cdr global))
 			(process-globals-info (cdr global)
 					      (->namestring (car global))
 					      get-package)))
@@ -472,7 +509,8 @@ USA.
 	 (lambda (package description)
 	   (let ((parent
 		  (let ((parent-name (package-description/parent description)))
-		    (and (not (eq? parent-name 'NONE))
+		    (and parent-name
+			 (not (eq? parent-name 'none))
 			 (get-package parent-name #t)))))
 	     (set-package/parent! package parent)
 	     (if parent
@@ -508,7 +546,7 @@ USA.
 	(let loop
 	    ((package package)
 	     (ancestors (vector-ref desc 1)))
-	  (if (eq? 'UNKNOWN (package/parent package))
+	  (if (eq? 'unknown (package/parent package))
 	      (if (pair? ancestors)
 		  (let ((parent (get-package (car ancestors) #t)))
 		    (set-package/parent! package parent)
@@ -533,19 +571,23 @@ USA.
 			     (vector-ref entry 2))))
 		    (link! package name
 			   external-package external-name
-			   package #f)))
+			   package #f
+			   (and (fix:= (vector-length entry) 4)
+				(vector-ref entry 3)))))
 		exports)))
 	  ;; Imported bindings.
 	  (for-each-vector-element (vector-ref desc 4)
 	    (lambda (entry)
 	      (let ((external-package (get-package (vector-ref entry 1) #t))
-		    (external-name 
+		    (external-name
 		     (if (fix:= (vector-length entry) 2)
 			 (vector-ref entry 0)
 			 (vector-ref entry 2))))
 		(link! external-package external-name
 		       package (vector-ref entry 0)
-		       package #f)))))))))
+		       package #f
+		       (and (fix:= (vector-length entry) 4)
+			    (vector-ref entry 3)))))))))))
 
 (define (for-each-exported-name exports receiver)
   (for-each
@@ -572,9 +614,9 @@ USA.
 	     (package-loop (package/parent package))))))
 
 (define (name->package packages name)
-  (list-search-positive packages
-    (lambda (package)
-      (symbol-list=? name (package/name package)))))
+  (find (lambda (package)
+	  (symbol-list=? name (package/name package)))
+	packages))
 
 (define (process-package-description package description get-package)
   (let ((file-cases (package-description/file-cases description)))
@@ -584,21 +626,21 @@ USA.
 	      (append-map! (lambda (file-case)
 			     (append-map cdr (cdr file-case)))
 			   file-cases))))
-  (for-each (lambda (export)
-	      (let ((destination (get-package (car export) #t)))
-		(for-each (lambda (names)
-			    (link! package (cdr names)
-				   destination (car names)
-				   package #t))
-			  (cdr export))))
+  (for-each (lambda (name.exports)
+	      (let ((destination (get-package (car name.exports) #t)))
+		(for-each (lambda (export)
+			    (link! package (vector-ref export 1)
+				   destination (vector-ref export 0)
+				   package #t (vector-ref export 2)))
+			  (cdr name.exports))))
 	    (package-description/exports description))
-  (for-each (lambda (import)
-	      (let ((source (get-package (car import) #t)))
-		(for-each (lambda (names)
-			    (link! source (cdr names)
-				   package (car names)
-				   package #t))
-			  (cdr import))))
+  (for-each (lambda (name.imports)
+	      (let ((source (get-package (car name.imports) #t)))
+		(for-each (lambda (import)
+			    (link! source (vector-ref import 1)
+				   package (vector-ref import 0)
+				   package #t #f))
+			  (cdr name.imports))))
 	    (package-description/imports description)))
 
 (define primitive-package-name
@@ -621,7 +663,7 @@ USA.
 
 (define (link! source-package source-name
 	       destination-package destination-name
-	       owner-package new?)
+	       owner-package new? deprecated?)
   (let ((source-binding (intern-binding! source-package source-name #f)))
     (make-link source-binding
 	       (let ((binding
@@ -634,15 +676,15 @@ USA.
 			   (error "Attempt to reinsert binding:"
 				  destination-name destination-package))
 		       (if new? (set-binding/new?! binding #t))
+		       (if deprecated? (set-binding/deprecated?! binding #t))
 		       binding)
 		     (let ((binding
 			    (make-binding destination-package
 					  destination-name
 					  (binding/value-cell source-binding)
 					  new?)))
-		       (rb-tree/insert! (package/bindings destination-package)
-					destination-name
-					binding)
+		       (package/put-binding! destination-package binding)
+		       (if deprecated? (set-binding/deprecated?! binding #t))
 		       binding)))
 	       owner-package
 	       new?)))
@@ -658,12 +700,11 @@ USA.
 		 (let ((binding (make-binding package name value-cell new?)))
 		   (set-value-cell/source-binding! value-cell binding)
 		   binding))))
-	  (rb-tree/insert! (package/bindings package) name binding)
+	  (package/put-binding! package binding)
 	  binding))))
 
 (define (make-reference package name expression)
-  (let ((references (package/references package))
-	(add-reference!
+  (let ((add-reference!
 	 (lambda (reference)
 	   (set-reference/expressions!
 	    reference
@@ -671,13 +712,13 @@ USA.
 	   (set-expression/references!
 	    expression
 	    (cons reference (expression/references expression))))))
-    (let ((reference (rb-tree/lookup references name #f)))
+    (let ((reference (package/find-reference package name)))
       (if reference
 	  (begin
 	    (if (not (memq expression (reference/expressions reference)))
 		(add-reference! reference))
 	    reference)
 	  (let ((reference (%make-reference package name)))
-	    (rb-tree/insert! references name reference)
+	    (package/put-reference! package reference)
 	    (add-reference! reference)
 	    reference)))))
\ No newline at end of file
