@@ -2,8 +2,8 @@
 
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
-    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 Massachusetts
-    Institute of Technology
+    2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
+    2017, 2018 Massachusetts Institute of Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -158,9 +158,9 @@ Type `h' after entering dired for more info."
 	    buffer)))))
 
 (define (find-dired-buffer directory-spec)
-  (list-search-positive (buffer-list)
-    (lambda (buffer)
-      (equal? directory-spec (buffer-get buffer 'DIRED-DIRECTORY-SPEC)))))
+  (find (lambda (buffer)
+	  (equal? directory-spec (buffer-get buffer 'DIRED-DIRECTORY-SPEC)))
+	(buffer-list)))
 
 (define (dired-buffer-directory-spec buffer)
   (or (buffer-get buffer 'DIRED-DIRECTORY-SPEC)
